/* 

  floyd -- display load average via PC keyboard LED.

  Start date:
  Version 1.0       Sun May 16 17:59:03 1999
  Version 1.1       Wed May 19 17:20:15 1999
  Version 1.2alpha  Fri May 21 23:09:14 1999
  Version 1.2beta   Mon May 24 04:53:27 1999
  Version 1.3       Sun May 30 18:39:42 1999
  Version 2.0       Mon Jun 21 12:25:58 1999
  Version 2.0 (release)   Thu Dec  1 18:45:00 2005

  (c) 1999-2005 Seiichi Namba <sn@asahi-net.email.ne.jp>
  No warranty for this software.
 
  INSTALL: make all ; su -c 'make install'

 */

#include "floyd.h"
#ifdef PTHREAD
#include "voodoo.h"
#endif

/*
   Externs
 */

char progname[384];
#ifdef PTHREAD
#ifdef PROCTHREAD
char version[] = "2.0 Edition 1 with PTHREAD/PROCTHREAD";
#else
char version[] = "2.0 Edition 1 with PTHREAD";
#endif
#else
char version[] = "2.0 Edition 1";
#endif

int fdes = 0;			/* Default input file descriptor, 0: stdin */
char fname[80]="" ;		/* Default input filename: stdin */

float l1, l2, l3;               /* Load averages, only l1 is used */
time_t currenttime;

floyd_param gfprm = {		/* Grandmaster floyd_param */
  BLINKLED,			/* gfprm.led */

  0,				/* gfprm.debug */
  0,				/* gfprm.verbose */
  0,				/* gfprm.tfloyd */
  0,				/* gfprm.force */

  0L,				/* gfprm.interval */
  MAXBLINK,			/* gfprm.maxblink */
  SLEEP,			/* gfprm.fixedsleep */
  MSEC100,			/* gfprm.ontime */

  &l1,				/* gfprm.loadavgp */

#ifdef BIGLOAD_SW
  BIGLOAD,			/* gfprm.bigloadsw */
#endif
  LOADAVG1,			/* gfprm.loadavg1 */
  UPDATECYCLE			/* gfprm.updatecycle */
#ifdef PACKETLOAD		/** packetload.c **/
  ,
  0,				/* gfprm.wywhopt */
  "",				/* gfprm.ifname: NULL string */
  -1,				/* gfprm.iftop: ifs offset in /proc/net/dev */
  -1				/* gfprm.fsize: size of /proc/net/dev */
#endif
#ifdef MORSE
  ,
  0				/* Enable MORSE mode in heavy load or not */
#endif
};

/*
   Statics variables & functions
   These `static' declarations are stronger than the function's
   below (I didn't know that...).
 */
/* Option parse-control variables */
#ifdef TEST_MGR			/* jobmgr tester wants this */
int sleepopt = 0;		/* -s<fixedsleep> : is specified or not. */
#else
static int sleepopt = 0;	/* -s<fixedsleep> : is specified or not. */
#endif
static int doargforked = 0;	/* -F(fork) or not: if it is last set to 2 */
#ifdef PTHREAD
static int jobregistered = 0;	/* -D(define) or not: if it is last set to 2 */
#endif

static void update(floyd_param *fpp);
static int  getoptarg(int ac, char **av, const char *p, char *form, void *val);
static void doarg(int ac, char **av);
static void proccheck(char *path);

#ifdef PTHREAD		/* jobmgr tester wants this */
#ifndef TEST_MGR
static
#endif
void registerone();
#endif

int do_fork;
int ledmodeworking;		/* Now gfprm.tfloyd is not referred directly */

/*

   OS/user/misc interfaces

 */

void showversion(FILE *fp) {
  fprintf(fp, "%s\n", version);
}

void usage(FILE *fp) {
  fprintf(fp,
"Usage: %s [-N|-C|-S] [tty] [-h|-H] [-l<msec>] [-m<msec>] [-u<sec>]\n\
             [-s<msec>]"
#ifdef BIGLOAD_SW
" [-b<L.A.>]"
#endif
#ifdef MORSE
" [-/+M]"
#endif
" [-a<1|2|3|5|15>] [-/+v] [-/+d]\n"
#ifdef PACKETLOAD
"             [-o<p|l>] [-/+E] [-p] [-i<ifs>] [-/+t]"
#else
"             [-/+t]"
#endif
#ifdef PTHREAD
" [-D] [-F]\n"
#else
" [-D]\n"
#endif
	  , progname);
}

#ifdef PTHREAD
 /*
    Potential bug1: global `fdes' is used here.  When fname/fdes is
    stored in floyd_param and job-independent (in future), they
    should be all reset on exit with signal.  Or Keep them in
    floyd_param even in current version ?

    Potential bug2: FIXED. gfprm.tfloyd is not checked anymore. Instead,
    ledmodeworking is checked. This is set in main() with find_ledmode()
    in jobmgr.c.
 */
void terminator(int sig) {
  /* Turn off all LED. */
  if (ledmodeworking) {
#if 0
    if (ioctl(fdes, KDSETLED, 0)) {
      perror("KDSETLED");
    }
#else
    ledalloff();
#endif
    /*close(fdes);*/ /* This raises KDSETLED: Bad file descriptor. not here, */
  }
  if (gfprm.verbose || doargforked)	/* doargforked: set by -D */
    fprintf(stderr, "%s: signal %d received [%d]\n", progname, sig, getpid());
#ifdef linux
  pthread_kill_other_threads_np();
#else
  Maybe you loose. Check out if you have pthread_kill_other_threads_np().
#endif
  close(fdes);			/* but here ? */
  exit(0);
}
#else
void terminator(int sig) {
  /* Turn off all LED. */
  if (ledmodeworking) {
#if 0
    if (ioctl(fdes, KDSETLED, 0)) {
      perror("KDSETLED");
    }
#else
    ledalloff();
#endif
  }
  if (gfprm.verbose || doargforked)	/* doargforked: set by -D */
    fprintf(stderr, "%s [%d]: signal %d received.\n", progname, getpid(), sig);
  close(fdes);
  exit(sig);
}
#endif

static struct leddef {
  char *name;
  int bit;
} leds[3] = {
  { "scroll", LED_SCR },
  { "num", LED_NUM },
  { "caps", LED_CAP }
};

char *ledname(floyd_param *fpp) {
  struct leddef *lp;
  if (fpp->tfloyd) return "ASCII";
  for (lp = leds; (lp - leds) < 3; lp++)
    if (fpp->led == lp->bit) return lp->name;
  return (char *)NULL;
}

/* Other function required for -p/-e mode ? */
void showinit(floyd_param *fpp) {
  fprintf(stderr,
"progname\t%10s\n\
pid\t\t%10d\n\
vc\t\t%10s\n\
led\t\t%10s\n"
#ifdef PACKETLOAD
"interface\t%10s\n\
wywhopt\t\t%10d\n"
#endif
"debug\t\t%10d\n\
tfloyd\t\t%10d\n\
loadavgp\t%10d\n\
blink@LA1.0\t%10.2f ms\n\
maxblink\t%10.2f ms\n\
updatecycle\t% 10d s\n"
#ifdef BIGLOAD_SW
"bigload switch\t%10.1f LA\n"
#endif
#ifdef MORSE
"morse \t\t%10d\n"
#endif
"sleep\t\t%10.2f ms\n\
do_fork\t\t%10d\n\
--\n"
,
          progname,
	  getpid(),
	  /* DON'T use straight fname here. See -f flag, and befloyd() */
	  /* fname,*/
	  (*fname ? fname : "stdin"),
          ledname(fpp),
#ifdef PACKETLOAD
          fpp->ifname,
	  fpp->wywhopt,
#endif
          fpp->debug,
          fpp->tfloyd,
          (fpp->loadavgp == &l1 ? 1 :
	   (fpp->loadavgp == &l2 ? 5 :
	    (fpp->loadavgp == &l3 ? 15 : -1))),
          (float)fpp->loadavg1 / (float)MSEC1,
          (float)fpp->maxblink / (float)MSEC1,
          fpp->updatecycle,
#ifdef BIGLOAD_SW
          fpp->bigloadsw,
#endif
#ifdef MORSE
          fpp->morse,
#endif
          (float)fpp->fixedsleep / (float)MSEC1,
	  do_fork);
}

/*

   Load average mode subroutines

 */

#if defined(PTHREAD) && defined(PROCTHREAD)
static pthread_mutex_t floyd_proc_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

static char rawbuffer[100];
static int len = -666;          /* A sinister value */
static inline void getproc(char *fn) {
  int fd;

#ifdef DEBUG
  fprintf(stderr, "%s [%d]: getproc(), len:%d.\n",
	  progname, getpid(), len);
#endif
  fd = open(fn, O_RDONLY);
#if defined(PTHREAD) && defined(PROCTHREAD)
  if (pthread_mutex_lock (&floyd_proc_mutex) != 0) {
    perror("Lock mutex"); exit(1);
  }
  len = read(fd, rawbuffer, sizeof(rawbuffer) - 1);
  close(fd);
  rawbuffer[len] = '\0';
  if (pthread_mutex_unlock (&floyd_proc_mutex) != 0) {
    perror("Unlock mutex"); exit(1);
  }
#else
  len = read(fd, rawbuffer, sizeof(rawbuffer) - 1);
  close(fd);
  rawbuffer[len] = '\0';
#endif
}

#if defined(PTHREAD) && defined(PROCTHREAD)
#define F_MINSLEEP_DIVISION 4
static long minsleep = -1;      /* Needs to be negative. */

void floyd_set_minsleep(ulong x) {
  ulong tem;

  if (pthread_mutex_lock (&floyd_proc_mutex) != 0) {
    perror("Lock mutex"); exit(1);
  }
  if (minsleep < 0)
    minsleep = x / F_MINSLEEP_DIVISION;
  else if (minsleep > (tem = x / F_MINSLEEP_DIVISION)) {
    minsleep = tem;
  }
  if (pthread_mutex_unlock (&floyd_proc_mutex) != 0) {
    perror("Unlock mutex"); exit(1);
  }
}

/*
   A thread routine to read /proc/loadavg
   Currently this is registered first prior to all the loadavg mode threads,
   and threads are fired up in FILO order, so I think this thread should
   have the last pid among the packetload threads, but it's not always. Why.

   Uses 1/4 of gfprm.updatecycle.
 */

void *floyd_getproc_loop(floyd_param *fpp) {

  if (fpp->debug || fpp->verbose)
    fprintf(stderr, "floyd_getproc_loop [%d]\n--\n", getpid());

  for (;;) {
#ifdef linux
    getproc("/proc/loadavg");
#else
    you lose;
#endif
#ifdef DEBUG
    /*fprintf(stderr, "FLOYD_GETPROC_LOOP: buf:%s\n", rawbuffer);*/
    fprintf(stderr, "FLOYD_GETPROC_LOOP: minsleep: %s\n", 
	    (minsleep < 0 ? "YET" : "DETERMINED"));
    fprintf(stderr, "FLOYD_GETPROC_LOOP: minsleep: %ld\n",
	    (minsleep < 0 ? SEC1 : minsleep));
#endif
    usleep(minsleep < 0 ? SEC1 : minsleep);
  }
  return (void *)0;		/* Just to be PTHREAD compliant */
}
#endif /* defined(PTHREAD) && defined(PROCTHREAD) */

static inline void getload(float *l1, float *l2, float *l3) {
  char *buf;

#if defined(PTHREAD) && defined(PROCTHREAD)
  /* In PTHREAD, getproc() is looped by an independent thread.
     so, NOP here.
     */
  if (len == -666) {      /* the procthread not working yet. */
    /* fprintf(stderr, "The procthread not working yet\n"); */
    *l1 = *l2 =*l3 = 0;
    return;
  } /* else fprintf(stderr, "The procthread WORKING !\n"); */
#else
#ifdef linux
  getproc("/proc/loadavg");
#else
  you lose;
#endif
#endif

  buf = rawbuffer;		/* good enough. won't write to the buffer */
#if defined(PTHREAD) && defined(PROCTHREAD) /**** NOT YET ****/
  if (pthread_mutex_lock (&floyd_proc_mutex) != 0) {
    perror("Lock mutex"); exit(1);
  }
  sscanf(buf, "%f%f%f", l1, l2, l3);
  if (pthread_mutex_unlock (&floyd_proc_mutex) != 0) {
    perror("Unlock mutex"); exit(1);
  }
#else
  sscanf(buf, "%f%f%f", l1, l2, l3);
#endif
}

static inline void update(floyd_param *fpp) {
  getload(&l1, &l2, &l3);

  if (*(fpp->loadavgp) == 0.0) fpp->interval = fpp->maxblink;
  else fpp->interval = (int)((float)fpp->loadavg1 / *(fpp->loadavgp));

  fpp->interval = (fpp->interval > fpp->maxblink) ?
    fpp->maxblink : fpp->interval;

  /* BIGLOAD fiddle: turn-on duration switched
     to SHORTON when (*(fpp->loadavgp) > fpp->bigloadsw).
     In Morse mode, NOP. */
#ifdef BIGLOAD_SW
#ifdef MORSE
  if (fpp->morse == 0)
    fpp->ontime = (*(fpp->loadavgp) > fpp->bigloadsw) ? SHORTON : LONGON;
#else
  fpp->ontime = (*(fpp->loadavgp) > fpp->bigloadsw) ? SHORTON : LONGON;
#endif
#endif

  if (fpp->debug)
    fprintf(stderr, "load:%f interval:%ld ontime:%ld\n",
	    *(fpp->loadavgp), fpp->interval, fpp->ontime);
}

/*

   Floyd upper layer controllers

 */

/* /proc fs check */
static void proccheck(char *path) {
  int tem = open(path, O_RDONLY); /* or access(2) is better ? */
  if (tem < 0) {		/* s.n. Fri May 28 15:38:23 1999 */
    perror(path);
    exit (1);
  }
  close(tem);
}
				
/* floyd loop
 */
void *floydloop(floyd_param *fpp) {
  register unchar l = fpp->led;
  register time_t lastupdate;
  register time_t ctt;		/* All temp */
  ulong tem;

  /* Do open-loop control without time() if
     (openloop = (int)(uupd / (fpp->ontime + offtime))) > F_OPENLOOP_THRES.
     Caveat: for() loop cycle is elongated, and MODULO operation is not
     usable in this method.
   */
#define F_OPENLOOP
#ifdef  F_OPENLOOP
#define F_OPENLOOP_THRESHOLD 8
  ulong uupd = fpp->updatecycle * SEC1;
  ulong offtime;
  int openloop;
#undef MODULO
#endif

  proccheck("/proc/loadavg");
  if (gfprm.verbose) showinit(fpp);

#if defined(PTHREAD) && defined(PROCTHREAD)
#if 1
  floyd_set_minsleep(fpp->fixedsleep);
#else
  floyd_set_minsleep((fpp->updatecycle) * SEC1);
#endif
#endif

#ifdef DEBUG
  fflush(stderr);
  fprintf(stderr, "(%d)floydloop ifname:%s [%5d:%5d]\n",
	  do_fork, fpp->ifname,  getpid(), (int)pthread_self());
  fflush(stderr);
#endif

  /* max-maxblink limited within updatecycle. Make sense ?
   */
  fpp->maxblink =
    fpp->maxblink > (tem = fpp->updatecycle * SEC1) ? tem : fpp->maxblink;
  update(fpp);
  currenttime = time(0);
  
  for (lastupdate = currenttime; ; ) {
#if !defined(F_OPENLOOP) && defined(MODULO)
    if (currenttime != (ctt = time(0))) {
      currenttime = ctt;
      if (currenttime % fpp->updatecycle == 0) update(fpp);
    }
#else  /* New code to avoid bignum division. */ 
    ctt = time(0);
    if (currenttime != ctt) {
      currenttime = ctt;
      if ((currenttime - lastupdate) >= fpp->updatecycle) {
        lastupdate = currenttime;
	update(fpp);
      }
    }
#endif

/* MORSE code without F_OPENLOOP 
 */
#ifdef MORSE
    if (fpp->morse && *(fpp->loadavgp) > fpp->bigloadsw) {

#ifdef MORSE_DEBUG
      fprintf(stderr,
"MORSE: *(fpp->loadavg):%f fpp->bigloadsw:%f int-load:%d compare-result:%d\n",
	      *(fpp->loadavgp), fpp->bigloadsw, (int)*(fpp->loadavgp),
	      (*(fpp->loadavgp) > fpp->bigloadsw) ? 1 : 0);
#endif

      Morse(l, (int)*(fpp->loadavgp));
      continue;
    }
#endif

/* MORSE code is tried to insert before blink().
   Not well tested, further I don't know if this is effective
   to cut %CPU, s.n. Fri Jan 21 18:58:42 2000
 */
#ifdef F_OPENLOOP
    offtime = fpp->interval + fpp->fixedsleep;
    openloop = (int)(uupd / (fpp->ontime + offtime));
    if (openloop <= F_OPENLOOP_THRESHOLD) {
	blink(l, fpp->ontime, offtime);
    } else do {
	/*fprintf(stderr, "openloop:%d\n", openloop);*/
	blink(l, fpp->ontime, offtime);
    } while (--openloop > 0);
#else
    blink(l, fpp->ontime, fpp->interval + fpp->fixedsleep);
    /*usleep(fpp->fixedsleep);*/
#endif
  }
  return (void *)0;		/* Just to be PTHREAD compliant */
}

/* wywh loop
 */
#ifdef PACKETLOAD
void *shineon(floyd_param *fpp) {

  proccheck("/proc/net/dev");
  if (gfprm.verbose) showinit(fpp);

#if defined(PTHREAD) && defined(PROCTHREAD)
  pload_set_minsleep(fpp->fixedsleep);
#endif

#ifdef DEBUG
  fflush(stderr);
  fprintf(stderr, "(%d)shineon ifname:%s [%5d:%5d]\n",
	  do_fork, fpp->ifname,  getpid(), (int)pthread_self());
  fflush(stderr);
#endif
  crazydiamond(fpp);
  /* crazydiamond() never returns, but just to be PTHREAD compliant */
  return (void *)0;
}
#endif

static void opendev(char *fname) {

  /* If any argument left, it is an input file (just one).
   */
  if (*fname) {
    if ((fdes = open(fname, O_RDWR)) < 0) {
      perror(fname);
      exit(1);
    }
  }
#if 0
  /* Why we shoudn't do this:
     Do not copy tty device filename here, or subsequent the same
     befloyd() -- opendev() (see -D option in doarg()) will do wrong
     thing at "if (*fname)" above.  Now showinit() knows how to
     generate filename, so this strcpy is totally unnecessary. */
  else strcpy(fname, "stdin");	/* default fdes is already 0 */
#endif
}

/* Do fancy stuffs if progname is such such.
 */
static void fancyname() {
  char *p;

  {				/* Move p to the top of file-name-nondir */
    p = strrchr(progname, '/');
    if (p) p++; else p = progname;
  }
  if (strcmp(p, "tfloyd") == 0) gfprm.tfloyd = 1; /* Names are exclusive */
#ifdef PACKETLOAD
  if (strcmp(p, "wywh") == 0) gfprm.wywhopt = 1;
#endif
}

/* 
   registerone() -- Register one floyd job
   Packetload specials are shredded into functions.
 */

#ifdef PACKETLOAD
void wywhfiddle() {

  if (*(gfprm.ifname)) {} /* Explicitly specified by -i, or -E|-w (forced), */
  else strcpy(gfprm.ifname, LISTEN_IFS); /* or default ("eth0") */
  if (gfprm.wywhopt && sleepopt == 0) /* If -s is not explicitly specified, */
    gfprm.fixedsleep = NAP;	/* take a default short NAP in loop. */
}
#endif

#ifdef PTHREAD			/* Now this is needed only in PTHREAD */
#ifndef TEST_MGR
static
#endif
void registerone() {

#ifdef PACKETLOAD
  wywhfiddle();
#endif

  /* not here, but from floydloop() or shineon() */
  /*if (gfprm.verbose) showinit();*/

#ifdef PACKETLOAD
  if (gfprm.wywhopt) {
    register_floyd_job(shineon);
  } else
#endif
  {
    register_floyd_job(floydloop);
  }
}
#endif

/*

   Command line parser

 */

/* Currently for number (read as float) and string processing only.
   Returns advanced step of arg processing.
   p should point the option character (l, m, u, b, S etc.).
 */
static int getoptarg(int ac, char **av, const char *p, char *form, void *val) {

  if (*(++p)) {                 /* -vl20.0 style, use (++p), not av[0]+2. */
    if (1 > sscanf(p, form, val)) goto quit;
    return 0;
  } else if (ac > 1) {          /* -vl 20.0 style, should args>=2 here */
    if (1 > sscanf(&(av[1][0]), form, val)) goto quit;
    return 1;
  } else {
    fprintf(stderr,
            "%s [%d]: option %s must have argument, Eugene,\n",
	    progname, getpid(), *av);
    exit(1);
  }
quit:
    fprintf(stderr,
            "%s [%d]: can't parse %s, Eugene,\n", progname, getpid(), *av);
    exit(1);
}

/* Option error handler used by doarg(). Mostly for existence of '-D'
 */
void doargexit() {

#ifndef PTHREAD /* Must be 1 when old '-D' is implemented also in PTHREAD. */
  if (doargforked) {
    fprintf(stderr, "%s [%d]: going to kill already detached child(ren).\n",
	    progname, getpid());
    killpg(0, SIGTERM);
  }
  /* not reached here */
#endif
  exit(1);
}

/* doarg(): scans all command line arguments.
 */
static void doarg(int ac, char **av) {

  register char *p;
  register char thisopt = '\0';	/* Just for relaying to prevopt */
  register char prevopt = '\0';
  float f;
  char flag = 0;

  /* A string containing -/+X possible `binary' option characters.
     You must update this if a new `binary' type option is added */
#ifdef PACKETLOAD
  char plusallowed[] = "vdtfpwE";
#else
  char plusallowed[] = "vdtf";
#endif

  for (ac--, av++; ac > 0 ; ac--, av++) {
    switch (**av) {
      case '-': flag = 1; break; /* enable binary */
      case '+': flag = 0; break; /* disable binary */
      default: flag = -1; break; /* perhaps fname arg */
    }

    if (flag >= 0) {
      p = *av + 1;              /* p moves within this single arg */
    cue:
      prevopt = thisopt;	/* Previous opt */
      thisopt = *p;
      /* if flag is `+', do `binary' type check for the optchar */
      if (flag == 0 && strchr(plusallowed, *p) == NULL)	goto parserr;

      switch (*p) {
	/* '\0' Means '-' only. AND this MUST be checked here, then break:
	   I don't want ++p anymore. */
	case '\0':
	  fdes = 0;		/* NOP might sound almost okay, but */
	  break;
	/* `binary' type, supporting +/-
	 */
	case 'v': gfprm.verbose= flag; if (*(++p)) goto cue; break;
	case 'd': gfprm.debug = flag;  if (*(++p)) goto cue; break;
	case 't': gfprm.tfloyd = flag; if (*(++p)) goto cue; break;
#ifdef PACKETLOAD
	case 'p': gfprm.wywhopt = flag; if (*(++p)) goto cue; break;
	case 'w':		/* WYWH mode */
	case 'E':		/* Now "-pi eth0" equivalent ! */
	  gfprm.wywhopt = flag;
	  strcpy(gfprm.ifname, LISTEN_IFS); /* eth0 is forced. */
	  if (*(++p)) goto cue; break;
#endif
#ifdef MORSE
	/* Still hidden option */
	case 'M': gfprm.morse = flag; if (*(++p)) goto cue; break;
#endif
	/* `exclusive-each-other' type
	 */
	case 'N':
	  gfprm.led = LED_NUM;
	  if (*(++p)) goto cue;
	  break;
	case 'C':
	  gfprm.led = LED_CAP;
	  if (*(++p)) goto cue; 
	  break;
	case 'S':
	  gfprm.led = LED_SCR; 
	  if (*(++p)) goto cue;
	  break;
	/* `number argument' type
	   All the time val except updatecycle must be specified in `msec'.
	   Pointer p never move in these tests (points relevant optchar).
	   */
	case 'l':
	  if (getoptarg(ac, av, p, "%f", &f) > 0) { ac--; av++; };
	  if (f <= 0) goto le0err; gfprm.loadavg1 = (float)(f * MSEC1);
	  break;
	case 'm':
	  if (getoptarg(ac, av, p, "%f", &f) > 0) { ac--; av++; };
	  if (f <= 0) goto le0err; gfprm.maxblink = (int)(f * MSEC1);
	  break;
	case 'u':
	  if (getoptarg(ac, av, p, "%f", &f) > 0) { ac--; av++; };
	  if (f <= 0) goto le0err; gfprm.updatecycle = (int)f;
	  /* Must not be 0 (or FP exception occurs in % operation below) */
	  if (gfprm.updatecycle <= 0) {
	    gfprm.updatecycle = 1; /* Minimum 1 sec */
	    if (gfprm.verbose)
	      fprintf(stderr,
            "%s [%d]: (warning) updatecycle is set to 1 even if %s specified.\n",
		      progname, getpid(),
		      *av);	/* *av: "-vu0", or just "0" (when av++) */
	  }
	  break;
	case 's':		/* ZERO sleep is allowed (for debug) */
	  if (getoptarg(ac, av, p, "%f", &f) > 0) { ac--; av++; };
	  if (f < 0) goto lt0err; gfprm.fixedsleep = (long)(f * MSEC1);
	  sleepopt = 1;
	  break;
#ifdef BIGLOAD_SW
	case 'b':		/* Maybe removed in future */
	  if (getoptarg(ac, av, p, "%f", &(gfprm.bigloadsw)) > 0)
	    { ac--; av++; };
	  if (gfprm.bigloadsw <= 0) goto le0err;
	  break;
#endif
	/* Still hidden opt. Select l1, l2, or l3 as loadavg */
	case 'a':
	  {
	    if (getoptarg(ac, av, p, "%f", &f) > 0) { ac--; av++; };
	    switch ((int)f) {
	      case 1:
		gfprm.loadavgp = &l1;
		break;
	      case 2:
	      case 5:
		gfprm.loadavgp = &l2;
		break;
	      case 3:
	      case 15:
		gfprm.loadavgp = &l3;
		break;
	      default: goto parserr;
	    }
	  }
	  break;

	/* `string argument' type, currently hidden option
	 */
#ifdef PACKETLOAD
	/* `o'perating mode: Currently LA/wywh only at most, so,
	   unused in non-PACKETLOAD compilation. */
	case 'o':		
	  {
	    char opmode;
	    if (getoptarg(ac, av, p, "%c", &opmode) > 0) { ac--; av++; };
	    switch (opmode) {	/* scan 1st one char only */
	      case 'p':
		gfprm.wywhopt = 1; break;
	      case 'l':
		gfprm.wywhopt = 0; break;
	      default: goto parserr;
	    }
	  }
	  break;
	case 'i':		/* `i'nterface spec */
	  if (getoptarg(ac, av, p, "%s", &(gfprm.ifname)) > 0) { ac--; av++; };
	  break;
#endif
	/* A `special' command-sort-of */
#ifdef PTHREAD
	case '-':		/* -- is alias of -D  */
	case 'D':		/* `D'efines one floyd `job' */
	case 'V':		/* Now Voodoo option == -D currently */
	  jobregistered = (ac > 1 || *(p+1)) ? 1 : 2;
	  registerone();
	  if (*(++p)) goto cue;
	  break;

	/* Was -D until 2.0pre01 Edtion 13. */
	case 'F':		/* Just set do_fork flag in PTHREAD */
	  doargforked = (ac > 1 || *(p+1)) ? 1 : 2;
	  do_fork = 1;
	  /* eg., in case -DF is the last arg,
	     To keep the parent from doing the job registered by the last 'F'
	     "-D /dev/tty -F" is also treated by this.  Does it make sense ?
	     */
	  if (doargforked == 2 &&
	      /* jobregistered && */
	      (prevopt == 'V' || prevopt == '-')) {
	    jobregistered = 2;
	  }
#else /* non-PTHREAD */
	case '-':		/* -- is alias of -D  */
	case 'D':		/* Works as detach immediately NOW ! */
	  doargforked = (ac > 1 || *(p+1)) ? 1 : 2;
	  do_fork = 1;
	  /* Now non-PTHREAD version is the same as 1.3.x:
	     -D works as `do-fork-now' */
	  ledmodeworking = find_ledmode();
	  /* registerone(); */      /* Not required anymore for do_job() */
	  do_job(PARENT_RETURN);
	  /* Parent exits if 'D' is the last optchar */
	  if (doargforked == 2) exit(0);
#endif
	  if (*(++p)) goto cue;
	  break;
	/* `exit' type */
	case 'H':
	  showversion(stdout); exit(0);
	case 'h':
	  usage(stdout); exit(0);
	default:
	  /*fprintf(stderr, "%s: No such option: %s\n", progname, *av);*/
	  /* How can put an echo with the \07 */
	  fprintf(stderr,
		  "\n\07careful with that argument %s, Eugene,\n\n", *av);
	  goto parserr;
      }
    } else {
      strcpy(fname, *av);	/* Or, it is VC filename */
      opendev(fname);
    }
    /*printf("doarg:%d led:%d arg:%s\n", wywhopt, led, *av);*/
  }
  return;
parserr:
  fprintf(stderr, "%s [%d]: can't parse %s.\n", progname, getpid(), *av);
  usage(stderr);
  doargexit();
le0err:
  fprintf(stderr, "%s [%d]: argument for -%c must be larger than 0.\n",
	  progname, getpid(), *p);
  doargexit();
lt0err:
  fprintf(stderr, "%s [%d]: argument for -%c must be larger or equal to 0.\n",
	  progname, getpid(), *p);
  doargexit();
}

/* Main
 */
#ifdef FLOYD
void main(int ac, char **av) {

  strcpy(progname, *av);
  fancyname();

  signal(SIGINT, terminator);
  signal(SIGHUP, terminator);
  signal(SIGTERM, terminator);

  doarg(ac, av);

  ledmodeworking = find_ledmode();

  /* If VC is not available, make an error-exit, because LED error after
     launching thread dumps core (not always) in the PROCTHREAD version.
     fdes is surely okay here.
     Perhaps PTHREAD only requires this. */
#if defined(PTHREAD) && defined(PROCTHREAD)
  if (ledmodeworking)
    ledalloff();			/* aka. ledcheck() */
#endif

#ifdef PTHREAD
#ifdef DEBUG
  fprintf(stderr,"jobregistered:%d\n", jobregistered);
#endif
  /* Good old -D is not implemented yet. */
  switch(jobregistered) {
    /* 
       jobregistered == 1 means user put -V in the middle: 
       the last option char is not -V.  */
    case 1:
#ifdef DEBUG
      fprintf(stderr,"do_fork:%d pid:%d\n", do_fork, getpid());
#endif
      launch_all(PARENT_EXIT, NO_JOIN); /* Just create thread */
#ifdef DEBUG
      fprintf(stderr,"afterlaunch:%d pid:%d\n", do_fork, getpid());
#endif
      /* registerone(); */      /* Not required anymore for do_job() */
      do_job(PARENT_EXIT);	/* Goes into function loop: with no thread */
      break;

    /* User put -V at the end. Do the job(s) as thread(s). */
    case 2:
      launch_all(PARENT_EXIT, DO_JOIN); /* This calls voodoo_join() */
      break;

    /* No -V specified: Do the one job with Master process.
       registerone() is not called yet here. */
    case 0:
      /* registerone(); */      /* Not required anymore for do_job() */
      do_job(PARENT_EXIT);	/* Goes into loop: without threading */
  }
#else  /* PTHREAD */
  /* registerone(); */          /* Not required anymore for do_job() */
  do_job(PARENT_EXIT);		/* Goes into loop: without threading */
#endif /* PTHREAD */
}
#endif /* FLOYD */
/*
  M-x compile for debug version, otherewise, just issue "make" from shell.

"make DEBUGDEF=-DDEBUG floyd"

  local variables:
  compile-command: "make clean voodoo"
  tab-width: 8
  c-indent-level: 2
  c-continued-statement-offset:  2
  c-brace-offset: 0
  c-argdecl-indent: 0
  c-label-offset: 2
  end:
  ;; Some of offsets seems like not working in 19.34/20.x. Use this:
  eval: (progn (c-set-offset 'label 0 nil) (c-set-offset 'case-label '+ nil))
  ;;tab-width: 2 (this is not what I want)
  ;; This is not auto-eval'ed:, authour likes eval this manually.
  (untabify (point-min) (point-max))
 */
