/*

  howmanymoretimes.c -- Includes howmanymoretimes()

  This file can be compiled as a stand alone program to test
  if your keyboard is compatible with floyd. Try

  $ cc howmanymoretimes.c ; ./a.out

  to test your keyboard. It loops 4 seconds, and dumps some
  debugging messages. If it does not raise any ^G nor warning
  message, probably your keyboard is floyd-compatible. The
  ./a.out may take one arg [N|C|S], to specify an LED to use in
  the test.

  howmanymoretimes.c is conceived to provide a
  is-keyboard-connected-error check in floyd (not yet
  implanted) by a function howmanymoretimes():


  NAME		howmanymoretimes()

  SYNOPSIS	time_t howmanymoretimes(unsigned long *diff)

  DESCRIPTION	Return time(2)-compatible time_t.
		Also stores time diff from the last call in msec as *diff.

		howmanymoretimes() are conceived for a is-keyboard-connected?
		test for floyd(1), but not implemented in floyd(1) yet.
		

  HISTORY
  Started:
  Thu Jun 17 23:00:28 1999  Seiichi Namba  <sn@asahi-net.email.ne.jp>
  
  Thu Jun 24 12:21:10 1999  main() works as test program. Takes one optional
                            arg [N|C|S|n|c|s] to specify the LED. Default is
                            NumLock LED.
                            To compile: M-x compile RET
			    then ./a.out [N|C|S|n|c|s]
  Fri Jul 30 10:36:52 1999  gettimeofday() support, but default is still ftime().
                            Current gettimeofday() code needs *1000 operation
			    to get millisecond unit.

An example uptime/ps output with an incompatible keyboard (actually it
is the MS `Natural' Elite in combination with GIGABYTE 6BA
motherboard), in which %CPU is (always) high, despite of low load
average 0.00. If there is eth0/ul0 traffic, the corresponding floyd's
%CPU goes up to 30-40%.  Spurious chatter are generated in that situation.

$ uptime
 11:26am  up 6 min,  1 user,  load average: 0.00, 0.13, 0.08
$ psgrep floyd
USER       PID %CPU %MEM  SIZE   RSS TTY STAT START   TIME COMMAND
root       198  4.2  0.2   824   300  ?  S    11:21   0:13 /usr/local/bin/floyd -v -C 
root       199  0.0  0.2   824   288  ?  S    11:21   0:00 /usr/local/bin/floyd -v -ES 
root       200  0.0  0.2   824   284  ?  S    11:21   0:00 /usr/local/bin/floyd -v -pi ul0 -N 

In this case, ./a.out from this program generates:
"^GB:Too slow IOCTL (31). Keyboard disconnected / IO broken ?"

 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/kd.h>

#include <time.h>

/* ftime() is default for a while */
#ifdef HAVE_GETTIMEOFDAY
#else
#undef HAVE_GETTIMEOFDAY
#endif

#ifdef HAVE_GETTIMEOFDAY
/* gettimeofday() */
#include <sys/time.h>
#else
#include <sys/timeb.h>
#endif

#ifndef BLINKLED
/* #define BLINKLED LED_SCR */  /* ScrollLock */
#define BLINKLED LED_NUM  /* NumLock */
/*#define BLINKLED LED_CAP*/        /* CapsLock, Default */
#endif

/* vcfd arg added */
static inline int ledctrl(unchar led, int onoff) {
#define DO_ON
  char oleds;
#ifdef DO_ON
  char nleds;
  char nval, ndef;
#endif

  if (ioctl(0, KDGETLED, &oleds)) {
    perror("KDGETLED");
    fprintf(stderr,
	    "Error reading current led setting. Maybe stdin is not a VT?\n");
    exit(1);
  }

#ifndef DO_ON
  return oleds;
#endif

#ifdef DO_ON
  nval = ndef = 0;
  ndef |= led;
  nval |= onoff ? led : 0;

  nleds = (oleds & ~ndef) | nval;

  if (ioctl(0, KDSETLED, nleds)) {
    perror("KDSETLED");
    exit(1);
  }
  return oleds;
#endif
}


/*
  NAME		howmanymoretimes()

  SYNOPSIS	time_t howmanymoretimes(unsigned long *diff)

  DESCRIPTION	Return time(2)-compatible time_t.
		Also stores time diff from the last call in msec in *ct.
 */
#ifdef HAVE_GETTIMEOFDAY
time_t howmanymoretimes(unsigned long *diff) {
  static struct timeval tval0;
  static int init = -1;

  struct timeval tval;
  time_t ct;			/* Current time */

  gettimeofday(&tval, (void *)0);
  ct = tval.tv_sec;

  if (init < 0) {		/* Then copy tval to tval0 (see below) */
    init = 1;
    if (diff) *diff = 0L;
  } else if (diff) {

    *diff = (tval0.tv_sec == tval.tv_sec) ?
      (unsigned long)tval.tv_usec - tval0.tv_usec                :
      (unsigned long)((1000000L * (tval.tv_sec - tval0.tv_sec - 1)) +
		(tval.tv_usec + (1000000L - tval0.tv_usec)))
      ;
    *diff /= 1000;
  }

  tval0 = tval;
  return tval.tv_sec;
}
#else /*HAVE_GETTIMEOFDAY*/
time_t howmanymoretimes(unsigned long *diff) {
  static struct timeb tb0;
  static int init = -1;

  struct timeb tb;
  time_t ct;			/* Current time */

  ftime(&tb);
  ct = tb.time;

  if (init < 0) {		/* Then copy tb to tb0 (see below) */
    init = 1;
    if (diff) *diff = 0L;
  } else if (diff) {

    *diff = (tb0.time == tb.time) ?
      (unsigned long)tb.millitm - tb0.millitm                :
      (unsigned long)((1000L * (tb.time - tb0.time -1)) +
		(tb.millitm + (1000L - tb0.millitm)))
      ;
  }

  /*tb0.time = tb.time;
    tb0.millitm = tb.millitm;*/
  tb0 = tb;

  return tb.time;
}
#endif /*HAVE_GETTIMEOFDAY*/

/* 
   printf() is slow ?  This
          printf("howmanymoretimes0:\n");
   sort of call takes 5ms ? (really ?  ain't it a howmanymoretimes() bug ?)
   in Celeron 300A.
*/
long fixedsleep=200000L;	/* in usec */
#define TESTPERIOD 4
void main(int ac, char **av) {
  int r, count, warning = 0;
  int onoff = 0;

  time_t start, currenttime;
  unsigned long delta;
  unchar led = BLINKLED;

  count = 0;

  if (ac > 1) {
    led = av[1][0];
    switch(led) {
      case 'n':
      case 'N':
	led = LED_NUM;
	break;
      case 'c':
      case 'C':
	led = LED_CAP;
	break;
      case 's':
      case 'S':
	led = LED_SCR;
	break;
      default:
	fprintf(stderr, "No such LED: %c\n", led);
	exit(1);
    }
  }

  start = howmanymoretimes(0); /* INIT here */
  printf("ctime: %s", ctime(&start));
  
  for (; ;count++) {
    /* The TEST. This is it.
       If the delay is thought to be large comparing to LOOP cycle,
       the keyboard might be disconnected. */
#if 1				/* Test A, Loop speed */
    currenttime = howmanymoretimes(&delta);
    printf("howmanymoretimes01: %ld\n", delta);

    if (delta > (long)(fixedsleep * 2 / 1000)) {
      printf("\07A:Too slow LOOP (%ld). Keyboard disconnected ?\n\n", delta);
      warning = 1;
    }
#endif

#if 1
    /* Only for this test prog. */
    if ((currenttime - start) > TESTPERIOD) break;
#endif

    r = ledctrl(led, onoff);
#ifdef DO_ON	
    onoff = onoff ? 0 : 1;
#endif
    printf("howmanymoretimes02: %d:%d\n", onoff, r);

    /* Perhaps this test should not be used to detect if keyboard is
       disconnected, because ioctl() return speed
       does not seem like depending on KB disconnection.

       However, some broken keyboards, or keyboard+motherboard combinations
       cannot pass this test. */
#if 1				/* Test B, IOCTL() speed */
    howmanymoretimes(&delta);
    printf("howmanymoretimes03: %ld\n\n", delta);
    if (delta > 20) { /* Passes good KSETLED is fast, even KB disconnected. */
      printf(
"\07B:Too slow IOCTL (%ld). Keyboard disconnected / IO broken ?\n\n", delta);
      warning = 1;
    }
#endif

#if 1
    usleep(fixedsleep);
#endif
  }

  printf("\nKeyboard test completed (%d seconds) %s", TESTPERIOD,
	 warning ?
	 "bug,\nas warned above, something is wrong on your KB...\n" :
	 "successfully.\n"
	 );

  printf("ctime: %s", ctime(&currenttime));

  ledctrl(led, 0);
  exit(warning);
}

/*
local variables:
 compile-command: "cc howmanymoretimes.c"
end:
*/
