/*
    voodoo.c
    Sun Jun  6 21:30:48 1999

    A pthread interface.

SYNOPSYS
    #include "voodoo.h"
    pthnode *voodoo_create(void (*fun)(), void *arg)
    void voodoo_join(pthnode *pthlp)
DESCRIPTION
    voodoo_create fires up a thread function fun with arg.
    It creates a structure node pthnode 

        typedef struct pthnode {
          pthread_t th;
          void *prev;
        } pthnode;

    and store a pthread_t identifier for the created thread, and link
    it to a master linked list.  The top of the list is stored in
    static pointer in voodoo_create().  On success the current top of
    the linked list is returned.  If the internal
    malloc/pthread_create failed exits with exit(errno).

*/

#ifdef TEST
 #include <signal.h>
 #include <linux/kd.h>
#endif
#include "voodoo.h"

/*
   Call pthnode_create() with fun and arg
*/

pthnode *voodoo_create(void *(*fun)(), void *arg) {
  static pthnode *lastlp = (void *)0; /* Initvalue compiled in this func. */
  pthnode *pthp;
#ifdef DEBUG			/** DEBUG **/
  static int count = 0;
  fprintf(stderr, "voodoo_create:%d\n", ++count);
  fflush(stderr);
#endif

  if ((pthp = malloc(sizeof(pthnode))) == NULL) {
    perror("voodoo malloc"); exit(errno);
  }
  pthp->prev = lastlp;		/* link the previous node, (void *)0 is 1st */

  /* pthp->th is setup by this */
  if (pthread_create(&(pthp->th), NULL, fun, arg)) {
    perror("voodoo pthread_create"); exit(errno);
  }
  return (lastlp = pthp);
}

/*
   Free the linked list. Static.
   Arg pthlp specifies the top of the list.
 */

static void voodoo_free(pthnode *pthlp) {
  pthnode *pthp;
  /* In reverse order from new to old */
  /*  for (pthp = pthlp; pthp ; pthp = pthp->prev) { *//* This seems to work ? */
  for (pthp = pthlp; pthp ; ) {
    pthnode *p;
    p = pthp->prev;
    free((void *)pthp);
#ifdef TEST
    fprintf(stderr, "curr:%ld prev:%ld\n", pthp, p);/*debugger*/
#endif
    pthp = p;
  }
}

/*
   Waits for all the threads invoked.
   Arg pthlp specifies the top of the list.
 */

void voodoo_join(pthnode *pthlp) {
  pthnode *pthp;

  /* In reverse order from new to old */
  for (pthp = pthlp; pthp ; pthp = pthp->prev)
    pthread_join(pthp->th, pthp->th_ret);
  voodoo_free(pthlp);		/* Free the linked list, really NEEDE ? */
}


#ifdef TEST

int fdes = 0;			/* means stdin */

void voodoo_kill(int sig) {	/* Now this is in floyd.c */
  /* Turn off all LED. */
  if (ioctl(fdes, KDSETLED, 0)) {
    perror("KDSETLED");
  }
  fprintf(stderr, "%s: signal %d received [%d]\n", progname, sig, getpid());
#ifdef linux
  pthread_kill_other_threads_np();
#else
  Maybe you loose. Check out if you have pthread_kill_other_threads_np().
#endif
  exit(0);
}

void *
fun1(void *arg) {
  int i;
  for (i = 0; i < 10; i++) {
    printf("fun1 %s [%d] %d\n", arg, getpid(), i);
    usleep(500000L);
  }
}

void *
fun2(void *arg) {
  int i;
  for (i = 0; i < 10; i++) {
    printf("fun2 %s [%d] %d\n", arg, getpid(), i);
    usleep(500000L);
  }
}

char progname[] ="voodootest";

void main() {
  pthnode *pthlp;		/* Master pointer to the linked list */
  int i;

  signal(SIGINT, voodoo_kill);
  signal(SIGHUP, voodoo_kill);
  signal(SIGTERM, voodoo_kill);

  pthlp = (void *)0;

  /* This runs immediately. */
  pthlp = voodoo_create((void *)fun1, "fun1arg");
  pthlp = voodoo_create((void *)fun2, "fun2arg");

  /*
  for (i = 0; i < 3; i++) {
    printf("main[%d] %i\n", getpid(), i);
    sleep(1);
  }
  */
  printf("pre join [%d]\n", getpid());
  voodoo_join(pthlp);
}
#endif
/*
  ;; Creates test version
  local variables:
  compile-command: "cc -DFLOYD -DTEST -s -static \\
-I /hdawork/src/glibc2/glibc-2.0.7pre6/linuxthreads/sysdeps/pthread \\
-o voodootest voodoo.c \\
-L /hdawork/src/glibc2/glibc-2.0.7pre6/linuxthreads -lpthread \\
-L /hdawork/src/glibc2/glibc-2.0.7pre6 -lc"
  end:
*/
